{% extends '!layout.html' %}

{%- block linktags %}
    <link href="https://yt-project.org/doc/{{ pagename }}.html" rel="canonical" />
    {{ super() }}
{%- endblock %}

{%- block extrahead %}
    {{ super() }}
    <script type="text/javascript">

      var _gaq = _gaq || [];
      _gaq.push(['_setAccount', 'UA-391373-2']);
      _gaq.push(['_setDomainName', 'yt-project.org']);
      _gaq.push(['_setAllowHash', false]);
      _gaq.push(['_trackPageview']);

      (function() {
        var ga = document.createElement('script'); ga.type = 'text/javascript'; ga.async = true;
        ga.src = ('https:' == document.location.protocol ? 'https://ssl' : 'http://www') + '.google-analytics.com/ga.js';
        var s = document.getElementsByTagName('script')[0]; s.parentNode.insertBefore(ga, s);
      })();

    </script>
{%- endblock %}

{%- block footer %}
    <div class="footer">
    {%- if hasdoc('copyright') %}
      {% trans path=pathto('copyright'), copyright=copyright|e %}&copy; <a href="{{ path }}">Copyright</a> {{ copyright }}.{% endtrans %}
    {%- else %}
      {% trans copyright=copyright|e %}&copy; Copyright {{ copyright }}.{% endtrans %}
    {%- endif %}
    {%- if last_updated %}
      {% trans last_updated=last_updated|e %}Last updated on {{ last_updated }}.{% endtrans %}
    {%- endif %}
    {%- if show_sphinx %}
      {% trans sphinx_version=sphinx_version|e %}Created using <a href="http://www.sphinx-doc.org/en/master/">Sphinx</a> {{ sphinx_version }}.{% endtrans %}
    {%- endif %}
    </div>
{%- endblock %}

{# Custom CSS overrides #}
{% set bootswatch_css_custom = ['_static/custom.css'] %}
